@@ -45,7 +45,7 @@
 #include <sys/ipc.h>
 #include <sys/sem.h>
 
-#define MOD_NSS_VERSION AP_SERVER_BASEREVISION
+#define MOD_NSS_VERSION PACKAGE_VERSION
 
 /* NSPR headers */
 #include "nspr.h"
@@ -59,6 +59,18 @@
 #include <nss.h>
 #include <sslproto.h>
 
+/* Apache ships its autoconf-generated config.h which defines these
+ * as empty. We want the mod_nss version in our own config.h so
+ * undefine them to eliminate the build warnings.
+ */
+#undef PACKAGE_NAME
+#undef PACKAGE_VERSION
+#undef PACKAGE_STRING
+#undef PACKAGE_TARNAME
+#undef PACKAGE_URL
+#undef PACKAGE_BUGREPORT
+#include "config.h"
+
 /* The #ifdef macros are only defined AFTER including the above
  * therefore we cannot include these system files at the top  :-(
  */
@@ -135,7 +147,7 @@ ap_set_module_config(c->conn_config, &nss_module, val)
 #define SSL_SESSION_CACHE_TIMEOUT  100
 #endif
 
-#ifndef SSL3_SESSION_CACHE_TIMEOUT 
+#ifndef SSL3_SESSION_CACHE_TIMEOUT
 #define SSL3_SESSION_CACHE_TIMEOUT  86400
 #endif
 
@@ -143,6 +155,11 @@ ap_set_module_config(c->conn_config, &nss_module, val)
 #define SSL_SESSION_CACHE_SIZE     10000
 #endif
 
+/* Default setting for per-dir reneg buffer. */
+#ifndef DEFAULT_RENEG_BUFFER_SIZE
+#define DEFAULT_RENEG_BUFFER_SIZE (128 * 1024)
+#endif
+
 /*
  * Define the SSL options
  */
@@ -159,7 +176,7 @@ typedef int nss_opt_t;
 
 /*
  * Define the SSL requirement structure
- */ 
+ */
 typedef struct {
     char     *cpExpr;
     nss_expr *mpExpr;
@@ -214,7 +231,7 @@ typedef enum {
 typedef struct {
     PRFileDesc *ssl;
     const char *client_dn;
-    CERTCertificate *client_cert; 
+    CERTCertificate *client_cert;
     int is_proxy;
     int disabled;
     int non_nss_request;
@@ -265,7 +282,7 @@ typedef struct {
 
 typedef struct {
     SSLSrvConfigRec *sc; /* pointer back to server config */
-    
+
     char *cipherSuite;
 
     int as_server;
@@ -306,12 +323,15 @@ struct SSLSrvConfigRec {
     const char      *ocsp_name;
     BOOL             ocsp;
     BOOL             enabled;
+    BOOL             sni;
+    BOOL             strict_sni_vhost_check;
     BOOL             proxy_enabled;
     const char      *vhost_id;
     int              vhost_id_len;
     modnss_ctx_t    *server;
     modnss_ctx_t    *proxy;
     BOOL             proxy_ssl_check_peer_cn;
+    BOOL             session_tickets;
 };
 
 /*
@@ -328,12 +348,17 @@ typedef struct {
     const char         *szCipherSuite;
     nss_verify_t        nVerifyClient;
     const char         *szUserName;
+    apr_size_t          nRenegBufferSize;
 } SSLDirConfigRec;
 
 /*
  * for cipher definitions see nss_engine_cipher.h
  */
 
+/* pool and hash to store ServerName and NSSNickname pairs for SNI */
+apr_pool_t *mp;
+apr_hash_t *ht;
+
 /* Compatibility between Apache 2.0.x and 2.2.x. The numeric version of
  * the version first appeared in Apache 2.0.56-dev. I picked 2.0.55 as it
  * is the last version without this define. This is used for more than just
@@ -351,7 +376,7 @@ typedef struct regex_t ap_regex_t;
 #define AP_REG_NOSUB REG_NOSUB
 #define AP_REG_ICASE REG_ICASE
 #endif
- 
+
 /*
  *  function prototypes
  */
@@ -366,6 +391,8 @@ void *nss_config_perdir_merge(apr_pool_t *p, void *basev, void *addv);
 void *nss_config_server_create(apr_pool_t *p, server_rec *s);
 void *nss_config_server_merge(apr_pool_t *p, void *basev, void *addv);
 const char *nss_cmd_NSSFIPS(cmd_parms *, void *, int);
+const char *nss_cmd_NSSSNI(cmd_parms *, void *, int);
+const char *nss_cmd_NSSStrictSNIVHostCheck(cmd_parms *, void *, int);
 const char *nss_cmd_NSSEngine(cmd_parms *, void *, int);
 const char *nss_cmd_NSSOCSP(cmd_parms *, void *, int);
 const char *nss_cmd_NSSOCSPDefaultResponder(cmd_parms *, void *, int);
@@ -391,10 +418,13 @@ const char *nss_cmd_NSSSessionCacheSize(cmd_parms *cmd, void *dcfg, const char *
 const char *nss_cmd_NSSPassPhraseDialog(cmd_parms *cmd, void *dcfg, const char *arg);
 const char *nss_cmd_NSSPassPhraseHelper(cmd_parms *cmd, void *dcfg, const char *arg);
 const char *nss_cmd_NSSRandomSeed(cmd_parms *, void *, const char *, const char *, const char *);
+const char *nss_cmd_NSSSessionTickets(cmd_parms *cmd, void *dcfg, int flag);
 const char *nss_cmd_NSSUserName(cmd_parms *cmd, void *dcfg, const char *arg);
 const char *nss_cmd_NSSOptions(cmd_parms *, void *, const char *);
 const char *nss_cmd_NSSRequireSSL(cmd_parms *cmd, void *dcfg);
 const char  *nss_cmd_NSSRequire(cmd_parms *, void *, const char *);
+const char *nss_cmd_NSSRenegBufferSize(cmd_parms *cmd, void *dcfg, const char *arg);
+
 
 const char *nss_cmd_NSSProxyEngine(cmd_parms *cmd, void *dcfg, int flag);
 const char *nss_cmd_NSSProxyProtocol(cmd_parms *, void *, const char *);
@@ -425,7 +455,7 @@ void         nss_var_log_config_register(apr_pool_t *p);
 
 APR_DECLARE_OPTIONAL_FN(char *, nss_var_lookup,
                         (apr_pool_t *, server_rec *,
-                         conn_rec *, request_rec *, 
+                         conn_rec *, request_rec *,
                          char *));
 
 /* An optional function which returns non-zero if the given connection
@@ -453,9 +483,12 @@ apr_file_t  *nss_util_ppopen(server_rec *, apr_pool_t *, const char *,
 void         nss_util_ppclose(server_rec *, apr_pool_t *, apr_file_t *);
 char        *nss_util_readfilter(server_rec *, apr_pool_t *, const char *,
                                  const char * const *);
+char *searchHashVhostbyNick(char *vhost_id);
+char *searchHashVhostbyNick_match(char *vhost_id);
+void addHashVhostNick(char *vhost_id, char *nickname);
 /* ssl_io_buffer_fill fills the setaside buffering of the HTTP request
  * to allow an SSL renegotiation to take place. */
-int          nss_io_buffer_fill(request_rec *r);
+int          nss_io_buffer_fill(request_rec *r, apr_size_t maxlen);
 
 int nss_rand_seed(server_rec *s, apr_pool_t *p, ssl_rsctx_t nCtx, char *prefix);
 
@@ -463,7 +496,11 @@ int nss_rand_seed(server_rec *s, apr_pool_t *p, ssl_rsctx_t nCtx, char *prefix);
 SECStatus nss_Init_Tokens(server_rec *s);
 
 /* Logging */
+#if AP_SERVER_MINORVERSION_NUMBER <= 2
+void nss_log_nss_error(const char *file, int line, int level, server_rec *s);
+#else
 void nss_log_nss_error(const char *file, int line, int module_index, int level, server_rec *s);
+#endif
 void nss_die(void);
 
 /* NSS callback */
