@@ -53,6 +53,7 @@ SSLModConfigRec *nss_config_global_create(server_rec *s)
     mc->pphrase_dialog_path         = NULL;
     mc->aRandSeed                   = apr_array_make(pool, 4,
                                                      sizeof(ssl_randseed_t));
+    mc->semid                       = 0;
 
     apr_pool_userdata_set(mc, SSL_MOD_CONFIG_KEY,
                           apr_pool_cleanup_null,
@@ -73,7 +74,6 @@ static void modnss_ctx_init(modnss_ctx_t *mctx)
 
     mctx->as_server           = PR_TRUE;
 
-    mctx->ssl2                = PR_FALSE;
     mctx->ssl3                = PR_FALSE;
     mctx->tls                 = PR_FALSE;
     mctx->tlsrollback         = PR_FALSE;
@@ -127,7 +127,7 @@ static void modnss_ctx_init_server(SSLSrvConfigRec *sc,
 static SSLSrvConfigRec *nss_config_server_new(apr_pool_t *p)
 {
     SSLSrvConfigRec *sc = apr_palloc(p, sizeof(*sc));
-    
+
     sc->mc                          = NULL;
     sc->ocsp                        = UNSET;
     sc->ocsp_default                = UNSET;
@@ -135,12 +135,15 @@ static SSLSrvConfigRec *nss_config_server_new(apr_pool_t *p)
     sc->ocsp_name                   = NULL;
     sc->fips                        = UNSET;
     sc->enabled                     = UNSET;
+    sc->sni                         = TRUE;
+    sc->strict_sni_vhost_check      = TRUE;
     sc->proxy_enabled               = UNSET;
     sc->vhost_id                    = NULL;  /* set during module init */
     sc->vhost_id_len                = 0;     /* set during module init */
     sc->proxy                       = NULL;
     sc->server                      = NULL;
     sc->proxy_ssl_check_peer_cn     = TRUE;
+    sc->session_tickets             = FALSE;
 
     modnss_ctx_init_proxy(sc, p);
 
@@ -214,8 +217,11 @@ void *nss_config_server_merge(apr_pool_t *p, void *basev, void *addv) {
     cfgMerge(ocsp_name, NULL);
     cfgMergeBool(fips);
     cfgMergeBool(enabled);
+    cfgMergeBool(sni);
+    cfgMergeBool(strict_sni_vhost_check);
     cfgMergeBool(proxy_enabled);
     cfgMergeBool(proxy_ssl_check_peer_cn);
+    cfgMergeBool(session_tickets);
 
     modnss_ctx_cfg_merge_proxy(base->proxy, add->proxy, mrg->proxy);
 
@@ -241,15 +247,17 @@ void *nss_config_perdir_create(apr_pool_t *p, char *dir) {
 
     dc->szUserName    = NULL;
 
+    dc->nRenegBufferSize = UNSET;
+
     return dc;
 }
- 
+
 const char *nss_cmd_NSSRequireSSL(cmd_parms *cmd, void *dcfg)
 {
     SSLDirConfigRec *dc = (SSLDirConfigRec *)dcfg;
 
     dc->bSSLRequired = TRUE;
- 
+
     return NULL;
 }
 
@@ -273,6 +281,23 @@ const char *nss_cmd_NSSRequire(cmd_parms *cmd,
     return NULL;
 }
 
+const char *nss_cmd_NSSRenegBufferSize(cmd_parms *cmd,
+                                       void *dcfg,
+                                       const char *arg)
+{
+    SSLDirConfigRec *dc = dcfg;
+    int val;
+
+    val = atoi(arg);
+    if (val < 0) {
+        return apr_pstrcat(cmd->pool, "Invalid size for NSSRenegBufferSize: ",
+                           arg, NULL);
+    }
+    dc->nRenegBufferSize = val;
+
+    return NULL;
+}
+
 void *nss_config_perdir_merge(apr_pool_t *p, void *basev, void *addv) {
     SSLDirConfigRec *base = (SSLDirConfigRec *)basev;
     SSLDirConfigRec *add  = (SSLDirConfigRec *)addv;
@@ -300,6 +325,8 @@ void *nss_config_perdir_merge(apr_pool_t *p, void *basev, void *addv) {
 
     cfgMergeString(szUserName);
 
+    cfgMergeInt(nRenegBufferSize);
+
     return mrg;
 }
 
@@ -308,16 +335,34 @@ const char *nss_cmd_NSSEngine(cmd_parms *cmd, void *dcfg, int flag)
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
 
     sc->enabled = flag ? TRUE : FALSE;
- 
+
     return NULL;
 }
 
 const char *nss_cmd_NSSFIPS(cmd_parms *cmd, void *dcfg, int flag)
 {
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
-    
+
     sc->fips = flag ? TRUE : FALSE;
- 
+
+    return NULL;
+}
+
+const char *nss_cmd_NSSSNI(cmd_parms *cmd, void *dcfg, int flag)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+
+    sc->sni = flag ? TRUE : FALSE;
+
+    return NULL;
+}
+
+const char *nss_cmd_NSSStrictSNIVHostCheck(cmd_parms *cmd, void *dcfg, int flag)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+
+    sc->strict_sni_vhost_check = flag ? TRUE : FALSE;
+
     return NULL;
 }
 
@@ -477,7 +522,7 @@ const char *nss_cmd_NSSRenegotiation(cmd_parms *cmd, void *dcfg, int flag)
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
 
     sc->server->enablerenegotiation = flag ? PR_TRUE : PR_FALSE;
- 
+
     return NULL;
 }
 
@@ -486,7 +531,7 @@ const char *nss_cmd_NSSRequireSafeNegotiation(cmd_parms *cmd, void *dcfg, int fl
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
 
     sc->server->requiresafenegotiation = flag ? PR_TRUE : PR_FALSE;
- 
+
     return NULL;
 }
 #endif
@@ -507,12 +552,12 @@ const char *nss_cmd_NSSECCNickname(cmd_parms *cmd,
 const char *nss_cmd_NSSProxyEngine(cmd_parms *cmd, void *dcfg, int flag)
 {
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
- 
+
     sc->proxy_enabled = flag ? TRUE : FALSE;
 
     return NULL;
 }
- 
+
 const char *nss_cmd_NSSProxyProtocol(cmd_parms *cmd,
                                      void *dcfg,
                                      const char *arg)
@@ -529,12 +574,12 @@ const char *nss_cmd_NSSProxyCipherSuite(cmd_parms *cmd,
                                         const char *arg)
 {
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
- 
+
     sc->proxy->auth.cipher_suite = arg;
- 
+
     return NULL;
 }
- 
+
 const char *nss_cmd_NSSProxyNickname(cmd_parms *cmd,
                                 void *dcfg,
                                 const char *arg)
@@ -562,7 +607,7 @@ const char *nss_cmd_NSSEnforceValidCerts(cmd_parms *cmd,
     SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
 
     sc->server->enforce = flag ? PR_TRUE : PR_FALSE;
- 
+
     return NULL;
 }
 
@@ -572,11 +617,10 @@ const char *nss_cmd_NSSSessionCacheTimeout(cmd_parms *cmd,
 {
     SSLModConfigRec *mc = myModConfig(cmd->server);
 
-    mc->session_cache_timeout = atoi(arg);
+    /* Deprecated. Store a value, if any, just to complain
+       about it later. */
 
-    if (mc->session_cache_timeout < 0) {
-        return "NSSSessionCacheTimeout: Invalid argument";
-    }
+    mc->session_cache_timeout = atoi(arg);
 
     return NULL;
 }
@@ -673,16 +717,16 @@ const char *nss_cmd_NSSRandomSeed(cmd_parms *cmd,
                                   const char *arg1,
                                   const char *arg2,
                                   const char *arg3)
-{   
+{
     SSLModConfigRec *mc = myModConfig(cmd->server);
     const char *err;
     ssl_randseed_t *seed;
     int arg2len = strlen(arg2);
-    
+
     if ((err = ap_check_cmd_context(cmd, GLOBAL_ONLY))) {
         return err;
     }
-    
+
     /* Only run through this once. Otherwise the random seed sources are
      * pushed into the array for each server start (and we are guaranteed 2) */
     if (mc->nInitCount >= 1) {
@@ -759,6 +803,14 @@ const char *nss_cmd_NSSRandomSeed(cmd_parms *cmd,
     return NULL;
 }
 
+const char *nss_cmd_NSSSessionTickets(cmd_parms *cmd,
+                                      void *dcfg, int flag)
+{
+    SSLSrvConfigRec *sc = mySrvConfig(cmd->server);
+    sc->session_tickets = flag ? PR_TRUE : PR_FALSE;
+    return NULL;
+}
+
 const char *nss_cmd_NSSUserName(cmd_parms *cmd, void *dcfg,
                                 const char *arg)
 {
@@ -772,10 +824,10 @@ const char *nss_cmd_NSSOptions(cmd_parms *cmd,
                                const char *arg)
 {
     SSLDirConfigRec *dc = (SSLDirConfigRec *)dcfg;
-    nss_opt_t opt;   
-    int first = TRUE; 
-    char action, *w; 
- 
+    nss_opt_t opt;
+    int first = TRUE;
+    char action, *w;
+
     while (*arg) {
         w = ap_getword_conf(cmd->pool, &arg);
         action = NUL;
