@@ -14,8 +14,13 @@
  */
 
 #include "mod_nss.h"
+#include "nss_engine_cipher.h"
 #include "apr_thread_proc.h"
+#include "apr_strings.h"
 #include "mpm_common.h"
+#if AP_SERVER_MINORVERSION_NUMBER <= 2
+#include "ap_mpm.h"
+#endif
 #include "secmod.h"
 #include "sslerr.h"
 #include "pk11func.h"
@@ -28,82 +33,26 @@ static SECStatus ownHandshakeCallback(PRFileDesc * socket, void *arg);
 static SECStatus NSSHandshakeCallback(PRFileDesc *socket, void *arg);
 static CERTCertificate* FindServerCertFromNickname(const char* name, const CERTCertList* clist);
 SECStatus nss_AuthCertificate(void *arg, PRFileDesc *socket, PRBool checksig, PRBool isServer);
+PRInt32 nssSSLSNISocketConfig(PRFileDesc *fd, const SECItem *sniNameArr, PRUint32 sniNameArrSize, void *arg);
 
 /*
  * Global variables defined in this file.
  */
 char* INTERNAL_TOKEN_NAME = "internal                         ";
 
-cipher_properties ciphers_def[ciphernum] =
-{
-    /* SSL2 cipher suites */
-    {"rc4", SSL_EN_RC4_128_WITH_MD5, 0, SSL2},
-    {"rc4export", SSL_EN_RC4_128_EXPORT40_WITH_MD5, 0, SSL2},
-    {"rc2", SSL_EN_RC2_128_CBC_WITH_MD5, 0, SSL2},
-    {"rc2export", SSL_EN_RC2_128_CBC_EXPORT40_WITH_MD5, 0, SSL2},
-    {"des", SSL_EN_DES_64_CBC_WITH_MD5, 0, SSL2},
-    {"desede3", SSL_EN_DES_192_EDE3_CBC_WITH_MD5, 0, SSL2},
-    /* SSL3/TLS cipher suites */
-    {"rsa_rc4_128_md5", SSL_RSA_WITH_RC4_128_MD5, 0, SSL3 | TLS},
-    {"rsa_rc4_128_sha", SSL_RSA_WITH_RC4_128_SHA, 0, SSL3 | TLS},
-    {"rsa_3des_sha", SSL_RSA_WITH_3DES_EDE_CBC_SHA, 0, SSL3 | TLS},
-    {"rsa_des_sha", SSL_RSA_WITH_DES_CBC_SHA, 0, SSL3 | TLS},
-    {"rsa_rc4_40_md5", SSL_RSA_EXPORT_WITH_RC4_40_MD5, 0, SSL3 | TLS},
-    {"rsa_rc2_40_md5", SSL_RSA_EXPORT_WITH_RC2_CBC_40_MD5, 0, SSL3 | TLS},
-    {"rsa_null_md5", SSL_RSA_WITH_NULL_MD5, 0, SSL3 | TLS},
-    {"rsa_null_sha", SSL_RSA_WITH_NULL_SHA, 0, SSL3 | TLS},
-    {"fips_3des_sha", SSL_RSA_FIPS_WITH_3DES_EDE_CBC_SHA, 0, SSL3 | TLS},
-    {"fips_des_sha", SSL_RSA_FIPS_WITH_DES_CBC_SHA, 0, SSL3 | TLS},
-    {"fortezza", SSL_FORTEZZA_DMS_WITH_FORTEZZA_CBC_SHA, 1, SSL3 | TLS},
-    {"fortezza_rc4_128_sha", SSL_FORTEZZA_DMS_WITH_RC4_128_SHA, 1, SSL3 | TLS},
-    {"fortezza_null", SSL_FORTEZZA_DMS_WITH_NULL_SHA, 1, SSL3 | TLS},
-    /* TLS 1.0: Exportable 56-bit Cipher Suites. */
-    {"rsa_des_56_sha", TLS_RSA_EXPORT1024_WITH_DES_CBC_SHA, 0, SSL3 | TLS},
-    {"rsa_rc4_56_sha", TLS_RSA_EXPORT1024_WITH_RC4_56_SHA, 0, SSL3 | TLS},
-    /* AES ciphers.*/
-    {"rsa_aes_128_sha", TLS_RSA_WITH_AES_128_CBC_SHA, 0, SSL3 | TLS},
-    {"rsa_aes_256_sha", TLS_RSA_WITH_AES_256_CBC_SHA, 0, SSL3 | TLS},
-#ifdef NSS_ENABLE_ECC
-    /* ECC ciphers.*/
-    {"ecdh_ecdsa_null_sha", TLS_ECDH_ECDSA_WITH_NULL_SHA, 0, TLS},
-    {"ecdh_ecdsa_rc4_128_sha", TLS_ECDH_ECDSA_WITH_RC4_128_SHA, 0, TLS},
-    {"ecdh_ecdsa_3des_sha", TLS_ECDH_ECDSA_WITH_3DES_EDE_CBC_SHA, 0, TLS},
-    {"ecdh_ecdsa_aes_128_sha", TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA, 0, TLS},
-    {"ecdh_ecdsa_aes_256_sha", TLS_ECDH_ECDSA_WITH_AES_256_CBC_SHA, 0, TLS},
-    {"ecdhe_ecdsa_null_sha", TLS_ECDHE_ECDSA_WITH_NULL_SHA, 0, TLS},
-    {"ecdhe_ecdsa_rc4_128_sha", TLS_ECDHE_ECDSA_WITH_RC4_128_SHA, 0, TLS},
-    {"ecdhe_ecdsa_3des_sha", TLS_ECDHE_ECDSA_WITH_3DES_EDE_CBC_SHA, 0, TLS},
-    {"ecdhe_ecdsa_aes_128_sha", TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA, 0, TLS},
-    {"ecdhe_ecdsa_aes_256_sha", TLS_ECDHE_ECDSA_WITH_AES_256_CBC_SHA, 0, TLS},
-    {"ecdh_rsa_null_sha", TLS_ECDH_RSA_WITH_NULL_SHA, 0, TLS},
-    {"ecdh_rsa_128_sha", TLS_ECDH_RSA_WITH_RC4_128_SHA, 0, TLS},
-    {"ecdh_rsa_3des_sha", TLS_ECDH_RSA_WITH_3DES_EDE_CBC_SHA, 0, TLS},
-    {"ecdh_rsa_aes_128_sha", TLS_ECDH_RSA_WITH_AES_128_CBC_SHA, 0, TLS},
-    {"ecdh_rsa_aes_256_sha", TLS_ECDH_RSA_WITH_AES_256_CBC_SHA, 0, TLS},
-    {"ecdhe_rsa_null", TLS_ECDHE_RSA_WITH_NULL_SHA, 0, TLS},
-    {"ecdhe_rsa_rc4_128_sha", TLS_ECDHE_RSA_WITH_RC4_128_SHA, 0, TLS},
-    {"ecdhe_rsa_3des_sha", TLS_ECDHE_RSA_WITH_3DES_EDE_CBC_SHA, 0, TLS},
-    {"ecdhe_rsa_aes_128_sha", TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA, 0, TLS},
-    {"ecdhe_rsa_aes_256_sha", TLS_ECDHE_RSA_WITH_AES_256_CBC_SHA, 0, TLS},
-    {"ecdh_anon_null_sha", TLS_ECDH_anon_WITH_NULL_SHA, 0, TLS},
-    {"ecdh_anon_rc4_128sha", TLS_ECDH_anon_WITH_RC4_128_SHA, 0, TLS},
-    {"ecdh_anon_3des_sha", TLS_ECDH_anon_WITH_3DES_EDE_CBC_SHA, 0, TLS},
-    {"ecdh_anon_aes_128_sha", TLS_ECDH_anon_WITH_AES_128_CBC_SHA, 0, TLS},
-    {"ecdh_anon_aes_256_sha", TLS_ECDH_anon_WITH_AES_256_CBC_SHA, 0, TLS},
-#endif
-};
+extern cipher_properties ciphers_def[];
 
 static char *version_components[] = {
     "SSL_VERSION_PRODUCT",
     "SSL_VERSION_INTERFACE",
     "SSL_VERSION_LIBRARY",
     NULL
-}; 
+};
 
 static char *nss_add_version_component(apr_pool_t *p,
                                        server_rec *s,
                                        char *name)
-{   
+{
     char *val = nss_var_lookup(p, s, NULL, NULL, name);
 
     if (val && *val) {
@@ -112,7 +61,7 @@ static char *nss_add_version_component(apr_pool_t *p,
 
     return val;
 }
- 
+
 static void nss_add_version_components(apr_pool_t *p,
                                        server_rec *s)
 {
@@ -135,16 +84,18 @@ static void nss_add_version_components(apr_pool_t *p,
  *  Initialize SSL library
  *
  */
-static void nss_init_SSLLibrary(server_rec *base_server)
+static void nss_init_SSLLibrary(server_rec *base_server, apr_pool_t *p)
 {
     SECStatus rv;
     SSLModConfigRec *mc = myModConfig(base_server);
-    SSLSrvConfigRec *sc; 
+    SSLSrvConfigRec *sc;
     char cwd[PATH_MAX];
     server_rec *s;
     int fipsenabled = FALSE;
     int ocspenabled = FALSE;
     int ocspdefault = FALSE;
+    int snienabled = FALSE;
+    char *dbdir = NULL;
     const char * ocspurl = NULL;
     const char * ocspname = NULL;
 
@@ -173,6 +124,10 @@ static void nss_init_SSLLibrary(server_rec *base_server)
                     return;
             }
         }
+
+        if (sc->sni == TRUE) {
+            snienabled = TRUE;
+        }
     }
 
     /* We need to be in the same directory as libnssckbi.so to load the
@@ -186,9 +141,15 @@ static void nss_init_SSLLibrary(server_rec *base_server)
             else
                 return;
     }
-    if (chdir(mc->pCertificateDatabase) != 0) {
+    if (strncasecmp(mc->pCertificateDatabase, "sql:", 4) == 0)
+        dbdir = (char *)mc->pCertificateDatabase + 4;
+    else
+        dbdir = (char *)mc->pCertificateDatabase;
+    if (chdir(dbdir) != 0) {
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
             "Unable to change directory to %s", mc->pCertificateDatabase);
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+            "Does the directory exist and do the permissions allow access?");
         if (mc->nInitCount == 1)
             nss_die();
         else
@@ -199,6 +160,8 @@ static void nss_init_SSLLibrary(server_rec *base_server)
     if (chdir(cwd) != 0) {
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
             "Unable to change directory to %s", cwd);
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+            "Does the directory exist and do the permissions allow access?");
         if (mc->nInitCount == 1)
             nss_die();
         else
@@ -207,14 +170,38 @@ static void nss_init_SSLLibrary(server_rec *base_server)
 
     /* Assuming everything is ok so far, check the cert database password(s). */
     if (rv != SECSuccess) {
+        apr_finfo_t finfo;
+        char keypath[1024];
+        int rv;
+        uid_t user_id;
+        gid_t gid;
+
+        user_id = ap_uname2id(mc->user);
+        gid = getegid();
+
         NSS_Shutdown();
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
             "NSS_Initialize failed. Certificate database: %s.", mc->pCertificateDatabase != NULL ? mc->pCertificateDatabase : "not set in configuration");
+
         nss_log_nss_error(APLOG_MARK, APLOG_ERR, base_server);
-        if (mc->nInitCount == 1)
-            nss_die();
-        else
-            return;
+        apr_snprintf(keypath, 1024, "%s/key3.db", mc->pCertificateDatabase);
+        if ((rv = apr_stat(&finfo, keypath, APR_FINFO_PROT | APR_FINFO_OWNER,
+             p)) == APR_SUCCESS) {
+            if (((user_id == finfo.user) &&
+                    (!(finfo.protection & APR_FPROT_UREAD))) ||
+                ((gid == finfo.group) &&
+                    (!(finfo.protection & APR_FPROT_GREAD))) ||
+                (!(finfo.protection & APR_FPROT_WREAD))
+               )
+            {
+                ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+                    "Server user lacks read access to NSS database.");
+            }
+        } else {
+            ap_log_error(APLOG_MARK, APLOG_ERR, 0, base_server,
+                "Does the NSS database exist?");
+        }
+        nss_die();
     }
 
     if (fipsenabled) {
@@ -267,7 +254,7 @@ static void nss_init_SSLLibrary(server_rec *base_server)
         /* We ensure that ocspname and ocspurl are not NULL above. */
         if (ocspdefault) {
             SECStatus sv;
- 
+
             sv = CERT_SetOCSPDefaultResponder(CERT_GetDefaultCertDB(),
                      ocspurl, ocspname);
 
@@ -294,7 +281,15 @@ static void nss_init_SSLLibrary(server_rec *base_server)
         }
     }
 
-    /* 
+    if (snienabled) {
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, base_server,
+            "SNI is enabled");
+    } else {
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, base_server,
+            "SNI is disabled");
+    }
+
+    /*
      * Seed the Pseudo Random Number Generator (PRNG)
      * only need ptemp here; nothing inside allocated from the pool
      * needs to live once we return from nss_rand_seed().
@@ -307,16 +302,18 @@ int nss_init_Module(apr_pool_t *p, apr_pool_t *plog,
                     server_rec *base_server)
 {
     SSLModConfigRec *mc = myModConfig(base_server);
-    SSLSrvConfigRec *sc; 
+    SSLSrvConfigRec *sc;
     server_rec *s;
     int sslenabled = FALSE;
     int fipsenabled = FALSE;
     int threaded = 0;
     struct semid_ds status;
+    char *split_vhost_id = NULL;
+    char *last1;
 
     mc->nInitCount++;
 
-    /* 
+    /*
      * Let us cleanup on restarts and exists
      */
     apr_pool_cleanup_register(p, base_server,
@@ -324,7 +321,7 @@ int nss_init_Module(apr_pool_t *p, apr_pool_t *plog,
                               apr_pool_cleanup_null);
 
     mc->ptemp = ptemp;
- 
+
     /*
      * Any init round fixes the global config
      */
@@ -333,8 +330,14 @@ int nss_init_Module(apr_pool_t *p, apr_pool_t *plog,
     /*
      * Fix up any global settings that aren't in the configuration
      */
-    if (mc->session_cache_timeout == UNSET) {
-        mc->session_cache_timeout = SSL_SESSION_CACHE_TIMEOUT;
+    if (mc->session_cache_timeout != UNSET) {
+        ap_log_error(APLOG_MARK, APLOG_WARNING, 0, base_server,
+            "NSSSessionCacheTimeout is deprecated. Ignoring.");
+
+        /* We still need to pass in a legal value to
+         * SSL_ConfigMPServerSIDCache() and SSL_ConfigServerSessionIDCache()
+         */
+        mc->session_cache_timeout = 0; /* use NSS default */
     }
 
     if (mc->ssl3_session_cache_timeout == UNSET) {
@@ -370,6 +373,14 @@ int nss_init_Module(apr_pool_t *p, apr_pool_t *plog,
         sc->vhost_id = nss_util_vhostid(p, s);
         sc->vhost_id_len = strlen(sc->vhost_id);
 
+        if (sc->sni && sc->server->nickname != NULL && sc->vhost_id != NULL) {
+            split_vhost_id = apr_strtok((char *)sc->vhost_id, ":", &last1);
+            ap_str_tolower(split_vhost_id);
+            addHashVhostNick(split_vhost_id, (char *)sc->server->nickname);
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                "SNI: %s -> %s", split_vhost_id, (char *)sc->server->nickname);
+	}
+
         /* Fix up stuff that may not have been set */
         if (sc->fips == UNSET) {
             sc->fips = FALSE;
@@ -504,7 +515,7 @@ int nss_init_Module(apr_pool_t *p, apr_pool_t *plog,
     PK11_ConfigurePKCS11(NULL,NULL,NULL, INTERNAL_TOKEN_NAME, NULL, NULL,NULL,NULL,8,1);
 
     ap_log_error(APLOG_MARK, APLOG_INFO, 0, base_server,
-        "Initializing SSL Session Cache of size %d. SSL2 timeout = %d, SSL3/TLS timeout = %d.", mc->session_cache_size, mc->session_cache_timeout, mc->ssl3_session_cache_timeout);
+        "Initializing SSL Session Cache of size %d. SSL3/TLS timeout = %d.", mc->session_cache_size, mc->ssl3_session_cache_timeout);
     ap_mpm_query(AP_MPMQ_MAX_THREADS, &threaded);
     if (!threaded)
         SSL_ConfigMPServerSIDCache(mc->session_cache_size, (PRUint32) mc->session_cache_timeout, (PRUint32) mc->ssl3_session_cache_timeout, NULL);
@@ -515,14 +526,14 @@ int nss_init_Module(apr_pool_t *p, apr_pool_t *plog,
     nss_io_layer_init();
 
     if (mc->nInitCount == 1) {
-        nss_init_SSLLibrary(base_server);
+        nss_init_SSLLibrary(base_server, mc->pPool);
         /*
          *  initialize servers
          */
         ap_log_error(APLOG_MARK, APLOG_INFO, 0, base_server,
                      "Init: Initializing (virtual) servers for SSL");
 
-        CERTCertList* clist = PK11_ListCerts(PK11CertListUser, NULL);
+        CERTCertList* clist = PK11_ListCerts(PK11CertListUserUnique, NULL);
 
         for (s = base_server; s; s = s->next) {
             sc = mySrvConfig(s);
@@ -616,13 +627,13 @@ static void nss_init_ctx_protocol(server_rec *s,
                                   apr_pool_t *ptemp,
                                   modnss_ctx_t *mctx)
 {
-    int ssl2, ssl3, tls, tls1_1, tls1_2;
+    int ssl3, tls, tls1_1, tls1_2;
     char *protocol_marker = NULL;
     char *lprotocols = NULL;
     SECStatus stat;
     SSLVersionRange enabledVersions;
 
-    ssl2 = ssl3 = tls = tls1_1 = tls1_2 = 0;
+    ssl3 = tls = tls1_1 = tls1_2 = 0;
 
     /*
      * Since this routine will be invoked individually for every thread
@@ -654,11 +665,7 @@ static void nss_init_ctx_protocol(server_rec *s,
             ap_str_tolower(lprotocols);
 
             if (strstr(lprotocols, "all") != NULL) {
-#ifdef WANT_SSL2
-                ssl2 = ssl3 = tls = tls1_1 = tls1_2 = 1;
-#else
                 ssl3 = tls = tls1_1 = tls1_2 = 1;
-#endif
             } else {
                 char *protocol_list = NULL;
                 char *saveptr = NULL;
@@ -669,16 +676,9 @@ static void nss_init_ctx_protocol(server_rec *s,
                     if (token == NULL) {
                         break;
                     } else if (strcmp(token, "sslv2") == 0) {
-#ifdef WANT_SSL2
-                        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
-                                     "%s:  Enabling SSL2",
-                                     protocol_marker);
-                        ssl2 = 1;
-#else
                         ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
                                      "%s:  SSL2 is not supported",
                                      protocol_marker);
-#endif
                     } else if (strcmp(token, "sslv3") == 0) {
                         ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
                                      "%s:  Enabling SSL3",
@@ -720,11 +720,7 @@ static void nss_init_ctx_protocol(server_rec *s,
 
     stat = SECSuccess;
 
-    if (ssl2 == 1) {
-        stat = SSL_OptionSet(mctx->model, SSL_ENABLE_SSL2, PR_TRUE);
-    } else {
-        stat = SSL_OptionSet(mctx->model, SSL_ENABLE_SSL2, PR_FALSE);
-    }
+    stat = SSL_OptionSet(mctx->model, SSL_ENABLE_SSL2, PR_FALSE);
 
     /* Set protocol version ranges:
      *
@@ -821,9 +817,18 @@ static void nss_init_ctx_protocol(server_rec *s,
         nss_die();
     }
 
-    mctx->ssl2 = ssl2;
     mctx->ssl3 = ssl3;
     mctx->tls = tls || tls1_1 || tls1_2;
+
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+        "%sabling TLS Session Tickets", mctx->sc->session_tickets == PR_TRUE ? "En" : "Dis");
+    if (SSL_OptionSet(mctx->model, SSL_ENABLE_SESSION_TICKETS,
+        mctx->sc->session_tickets) != SECSuccess) {
+        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
+                "Unable to configure TLS Session Tickets");
+        nss_log_nss_error(APLOG_MARK, APLOG_ERR, s);
+        nss_die();
+    }
 }
 
 static void nss_init_ctx_session_cache(server_rec *s,
@@ -881,21 +886,6 @@ static void nss_init_ctx_verify(server_rec *s,
     }
 }
 
-static int countciphers(PRBool cipher_state[ciphernum], int version) {
-    int ciphercount = 0;
-    int i;
-
-    for (i = 0; i < ciphernum; i++)
-    {
-        if ((cipher_state[i] == PR_TRUE) &&
-            (ciphers_def[i].version & version)) {
-            ciphercount++;
-        }
-    }
-
-    return ciphercount;
-}
-
 static void nss_init_ctx_cipher_suite(server_rec *s,
                                       apr_pool_t *p,
                                       apr_pool_t *ptemp,
@@ -903,14 +893,14 @@ static void nss_init_ctx_cipher_suite(server_rec *s,
 {
     PRBool cipher_state[ciphernum];
     PRBool fips_state[ciphernum];
-    const char *suite = mctx->auth.cipher_suite; 
+    const char *suite = mctx->auth.cipher_suite;
     char * object_type = NULL;
     char * cipher_suite_marker = NULL;
     char * ciphers;
     char * fipsciphers = NULL;
     int i;
- 
-    /* 
+
+    /*
      *  Configure SSL Cipher Suite
      */
     if (!suite) {
@@ -1037,22 +1027,14 @@ static void nss_init_ctx_cipher_suite(server_rec *s,
         }
     }
 
-    /* See if any ciphers have been enabled for a given protocol */
-    if (mctx->ssl2 && countciphers(cipher_state, SSL2) == 0) {
-        ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-            "%s:  SSL2 is enabled but no SSL2 ciphers are enabled.",
-            cipher_suite_marker);
-        nss_die();
-    }
-
-    if (mctx->ssl3 && countciphers(cipher_state, SSL3) == 0) {
+    if (mctx->ssl3 && countciphers(cipher_state, SSLV3) == 0) {
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
             "%s:  SSL3 is enabled but no SSL3 ciphers are enabled.",
             cipher_suite_marker);
         nss_die();
     }
 
-    if (mctx->tls && countciphers(cipher_state, TLS) == 0) {
+    if (mctx->tls && countciphers(cipher_state, TLSV1|TLSV1_2) == 0) {
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
             "%s:  TLS is enabled but no TLS ciphers are enabled.",
             cipher_suite_marker);
@@ -1061,7 +1043,7 @@ static void nss_init_ctx_cipher_suite(server_rec *s,
 
     /* Finally actually enable the selected ciphers */
     for (i=0; i<ciphernum;i++) {
-        SSL_CipherPrefSet(mctx->model, ciphers_def[i].num, cipher_state[i]);
+        SSL_CipherPrefSet(mctx->model, ciphers_def[i].num, cipher_state[i] == 1 ? PR_TRUE : PR_FALSE);
     }
 }
 
@@ -1085,17 +1067,17 @@ static void nss_init_server_check(server_rec *s,
 static void nss_init_ctx(server_rec *s,
                          apr_pool_t *p,
                          apr_pool_t *ptemp,
-                         modnss_ctx_t *mctx) 
+                         modnss_ctx_t *mctx)
 {
 
     nss_init_ctx_socket(s, p, ptemp, mctx);
 
     nss_init_ctx_protocol(s, p, ptemp, mctx);
-    
+
     nss_init_ctx_session_cache(s, p, ptemp, mctx);
-    
+
     nss_init_ctx_callbacks(s, p, ptemp, mctx);
-    
+
     nss_init_ctx_verify(s, p, ptemp, mctx);
 
     nss_init_ctx_cipher_suite(s, p, ptemp, mctx);
@@ -1107,13 +1089,20 @@ static void nss_init_certificate(server_rec *s, const char *nickname,
                                  SSLKEAType *KEAtype,
                                  PRFileDesc *model,
                                  int enforce,
+                                 int sni,
                                  const CERTCertList* clist)
 {
     SECCertTimeValidity certtimestatus;
     SECStatus secstatus;
 
     PK11SlotInfo* slot = NULL;
- 
+    CERTCertNicknames *certNickDNS = NULL;
+    char **nnptr = NULL;
+    int nn = 0;
+    apr_array_header_t *names = NULL;
+    apr_array_header_t *wild_names = NULL;
+    int i, j;
+
     if (nickname == NULL) {
         return;
     }
@@ -1143,13 +1132,13 @@ static void nss_init_certificate(server_rec *s, const char *nickname,
 
     if (strchr(nickname, ':'))
     {
-        char* token = strdup(nickname); 
+        char* token = strdup(nickname);
         char* colon = strchr(token, ':');
         if (colon) {
             *colon = 0;
             slot = PK11_FindSlotByName(token);
             if (!slot) {
-                /* 
+                /*
                  * Slot not found. This should never happen because we
                  * already found the cert.
                  */
@@ -1179,12 +1168,59 @@ static void nss_init_certificate(server_rec *s, const char *nickname,
 
     *KEAtype = NSS_FindCertKEAType(*servercert);
 
+    /* add ServerAlias entries to hash */
+    names = s->names;
+    if (names) {
+        char **name = (char **)names->elts;
+        for (i = 0; i < names->nelts; ++i) {
+            ap_str_tolower(name[i]);
+            addHashVhostNick(name[i], (char *)nickname);
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                "SNI ServerAlias: %s -> %s", name[i], nickname);
+        }
+    }
+
+    /* add ServerAlias entries with wildcards */
+    wild_names = s->wild_names;
+    if (wild_names) {
+        char **wild_name = (char **)wild_names->elts;
+        for (j = 0; j < wild_names->nelts; ++j) {
+            ap_str_tolower(wild_name[j]);
+            addHashVhostNick(wild_name[j], (char *)nickname);
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                "SNI wildcard: %s -> %s", wild_name[j], nickname);
+        }
+    }
+
+    /* get valid DNS names from certificate to hash */
+    certNickDNS = CERT_GetValidDNSPatternsFromCert(*servercert);
+
+    if (certNickDNS) {
+        nnptr = certNickDNS->nicknames;
+        nn = certNickDNS->numnicknames;
+
+        while ( nn > 0 ) {
+            ap_str_tolower(*nnptr);
+            addHashVhostNick(*nnptr, (char *)nickname);
+            nnptr++;
+            nn--;
+        }
+        PORT_FreeArena(certNickDNS->arena, PR_FALSE);
+    }
+
+    /* Subject/hostname check */
+    secstatus = CERT_VerifyCertName(*servercert, s->server_hostname);
+    if (secstatus != SECSuccess) {
+      char *cert_dns = CERT_GetCommonName(&(*servercert)->subject);
+      ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
+		   "Misconfiguration of certificate's CN and virtual name."
+		   " The certificate CN has %s. We expected %s as virtual"
+		   " name.", cert_dns, s->server_hostname);
+      PORT_Free(cert_dns);
+    }
     /*
-     * Check for certs that are expired or not yet valid and WARN about it
-     * no need to refuse working - the client gets a warning, but can work
-     * with the server we could also verify if the certificate is made out
-     * for the correct hostname but that would require a reverse DNS lookup
-     * for every virtual server - too expensive?
+     * Check for certs that are expired or not yet valid and WARN about it.
+     * No need to refuse working - the client gets a warning.
      */
 
     certtimestatus = CERT_CheckCertValidTimes(*servercert, PR_Now(), PR_FALSE);
@@ -1206,13 +1242,23 @@ static void nss_init_certificate(server_rec *s, const char *nickname,
             break;
     }
 
-    secstatus = SSL_ConfigSecureServer(model, *servercert, *serverkey, *KEAtype);
+    secstatus = SSL_ConfigSecureServer(model, *servercert, *serverkey,
+                                       *KEAtype);
     if (secstatus != SECSuccess) {
         ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
             "SSL error configuring server: '%s'", nickname);
         nss_log_nss_error(APLOG_MARK, APLOG_ERR, s);
         nss_die();
     }
+
+    if (PR_TRUE == sni) {
+        if (SSL_SNISocketConfigHook(model, (SSLSNISocketConfig) nssSSLSNISocketConfig, (void*) s) != SECSuccess) {
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                "SSL_SNISocketConfigHook failed");
+            nss_log_nss_error(APLOG_MARK, APLOG_ERR, s);
+            nss_die();
+        }
+    }
 }
 
 
@@ -1258,11 +1304,13 @@ static void nss_init_server_certs(server_rec *s,
 
         nss_init_certificate(s, mctx->nickname, &mctx->servercert,
                              &mctx->serverkey, &mctx->serverKEAType,
-                             mctx->model, mctx->enforce, clist);
+                             mctx->model, mctx->enforce, mctx->sc->sni,
+                             clist);
 #ifdef NSS_ENABLE_ECC
         nss_init_certificate(s, mctx->eccnickname, &mctx->eccservercert,
                              &mctx->eccserverkey, &mctx->eccserverKEAType,
-                             mctx->model, mctx->enforce, clist);
+                             mctx->model, mctx->enforce, mctx->sc->sni,
+                             clist);
 #endif
     }
 
@@ -1272,7 +1320,6 @@ static void nss_init_server_certs(server_rec *s,
             "Error setting PKCS11 pin argument: '%s'", mctx->nickname);
         nss_die();
     }
-    
     secstatus = (SECStatus)SSL_HandshakeCallback(mctx->model, (SSLHandshakeCallback)NSSHandshakeCallback, NULL);
     if (secstatus != SECSuccess)
     {
@@ -1281,6 +1328,7 @@ static void nss_init_server_certs(server_rec *s,
         nss_log_nss_error(APLOG_MARK, APLOG_ERR, s);
         nss_die();
     }
+
 }
 
 static void nss_init_proxy_ctx(server_rec *s,
@@ -1347,7 +1395,6 @@ void nss_init_Child(apr_pool_t *p, server_rec *base_server)
         /* If any servers have SSL, we want sslenabled set so we
          * can perform further initialization
          */
-
         if (sc->enabled == UNSET) {
             sc->enabled = FALSE;
         }
@@ -1374,20 +1421,21 @@ void nss_init_Child(apr_pool_t *p, server_rec *base_server)
         }
     }
 
-    nss_init_SSLLibrary(base_server);
+    nss_init_SSLLibrary(base_server, mc->pPool);
 
     /* Configure all virtual servers */
-    CERTCertList* clist = PK11_ListCerts(PK11CertListUser, NULL);
+    CERTCertList* clist = PK11_ListCerts(PK11CertListUserUnique, NULL);
     for (s = base_server; s; s = s->next) {
         sc = mySrvConfig(s);
-        if (sc->server->servercert == NULL && NSS_IsInitialized())
+        if (sc->server->servercert == NULL && NSS_IsInitialized()) {
             nss_init_ConfigureServer(s, p, mc->ptemp, sc, clist);
+        }
     }
     if (clist) {
         CERT_DestroyCertList(clist);
     }
 
-    /* 
+    /*
      * Let us cleanup on restarts and exits
      */
     apr_pool_cleanup_register(p, base_server,
@@ -1412,6 +1460,11 @@ apr_status_t nss_init_ModuleKill(void *data)
     if (mc->nInitCount == 1)
         nss_init_ChildKill(base_server);
 
+    if (mp) {
+        apr_pool_destroy(mp);
+        mp = NULL;
+    }
+
     /* NSS_Shutdown() gets called in nss_init_ChildKill */
     return APR_SUCCESS;
 }
@@ -1464,6 +1517,11 @@ apr_status_t nss_init_ChildKill(void *data)
         }
     }
 
+    if (mp) {
+        apr_pool_destroy(mp);
+        mp = NULL;
+    }
+
     if (shutdown) {
         /* Clear any client-side session cache data */
         SSL_ClearSessionCache();
@@ -1517,7 +1575,7 @@ SECStatus ownHandshakeCallback(PRFileDesc * socket, void *arg)
  */
 static PRBool
 cert_IsNewer(CERTCertificate *certa, CERTCertificate *certb)
-{ 
+{
     PRTime notBeforeA, notAfterA, notBeforeB, notAfterB, now;
     SECStatus rv;
     PRBool newerbefore, newerafter;
@@ -1605,11 +1663,11 @@ FindServerCertFromNickname(const char* name, const CERTCertList* clist)
              * Otherwise just return the cert if the nickname matches.
              */
             if (CERT_CheckCertUsage(cert, certUsageSSLServer) == SECSuccess) {
-                matchedUsage = 2; 
+                matchedUsage = 2;
             } else {
-                if (CERT_CheckCertUsage(cert, certUsageEmailRecipient) == SECSuccess) 
+                if (CERT_CheckCertUsage(cert, certUsageEmailRecipient) == SECSuccess)
                 {
-                    matchedUsage = 1; 
+                    matchedUsage = 1;
                 }
             }
 
@@ -1664,57 +1722,84 @@ SECStatus NSSHandshakeCallback(PRFileDesc *socket, void *arg)
     return SECSuccess;
 }
 
-int nss_parse_ciphers(server_rec *s, char *ciphers, PRBool cipher_list[ciphernum])
+/*
+ * Callback made during SSL request to see if SNI was requested and
+ * pair it with a configured nickname.
+ */
+PRInt32 nssSSLSNISocketConfig(PRFileDesc *fd, const SECItem *sniNameArr,
+           PRUint32 sniNameArrSize, void *arg)
 {
-    char * cipher;
-    PRBool found, active;
-    int i;
+    server_rec *s = (server_rec *)arg;
 
-    cipher = ciphers;
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+		 "nssSSLSNISocketConfig");
 
-    while (ciphers && (strlen(ciphers)))
-    {
-        while ((*cipher) && (isspace(*cipher)))
-           ++cipher;
+    void *pinArg;
+    CERTCertificate *cert = NULL;
+    SECKEYPrivateKey *privKey = NULL;
+    char *nickName = NULL;
+    char *vhost = NULL;
+    apr_pool_t *str_p;
 
-        switch(*cipher++)
-        {
-            case '+':
-                active = PR_TRUE;
-                break;
-            case '-':
-                active = PR_FALSE;
-                break;
-            default:
-                ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-                             "invalid cipher string %s. Format is +cipher1,-cipher2...", cipher - 1);
-            return -1;
-        }
+    PORT_Assert(fd && sniNameArr);
+    if (!fd || !sniNameArr) {
+        return SSL_SNI_SEND_ALERT;
+    }
 
-        if ((ciphers = strchr(cipher, ','))) {
-            *ciphers++ = '\0';
-        }
+    apr_pool_create(&str_p, NULL);
+    vhost = apr_pstrndup(str_p, (char *) sniNameArr->data, sniNameArr->len);
 
-        found = PR_FALSE;
+    /* rfc6125 - Checking of Traditional Domain Names */
+    ap_str_tolower(vhost);
 
-        for (i = 0; i < ciphernum; i++)
-        {
-            if (!strcasecmp(cipher, ciphers_def[i].name)) {
-                cipher_list[i] = active;
-                found = PR_TRUE;
-                break;
-            }
-        }
-
-        if (found == PR_FALSE) {
-            ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-                         "Unknown cipher %s", cipher);
-        }
-
-        if (ciphers) {
-            cipher = ciphers;
+    nickName = searchHashVhostbyNick(vhost);
+    if (nickName == NULL)  {
+        /* search for wildcard_names in serverAlises */
+        nickName = searchHashVhostbyNick_match(vhost);
+        if (nickName == NULL) {
+            ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+                "SNI: Search for %s failed. Unrecognized name.", vhost);
+            goto loser;
         }
     }
 
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,"SNI: Found nickname %s for vhost: %s", nickName, vhost);
+
+    pinArg = SSL_RevealPinArg(fd);
+
+    /* if pinArg is NULL, then we would not get the key and
+     * return an error status. */
+    cert = PK11_FindCertFromNickname(nickName, &pinArg);
+    if (cert == NULL) {
+        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+            "Failed to find certificate for nickname: %s", nickName);
+        goto loser;
+    }
+    privKey = PK11_FindKeyByAnyCert(cert, &pinArg);
+    if (privKey == NULL) {
+        goto loser;
+    }
+
+    SSLKEAType certKEA = NSS_FindCertKEAType(cert);
+
+    if (SSL_ConfigSecureServer(fd, cert, privKey, certKEA) != SECSuccess) {
+        goto loser; /* Send alert */
+    }
+    ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s,
+        "SNI: Successfully paired vhost %s with nickname: %s", vhost, nickName);
+
+    apr_pool_destroy(str_p);
+    SECKEY_DestroyPrivateKey(privKey);
+    CERT_DestroyCertificate(cert);
+
     return 0;
+
+loser:
+    if (privKey)
+        SECKEY_DestroyPrivateKey(privKey);
+    if (cert)
+        CERT_DestroyCertificate(cert);
+    apr_pool_destroy(str_p);
+
+    return SSL_SNI_SEND_ALERT;
 }
