@@ -201,7 +201,11 @@ char *nss_var_lookup(apr_pool_t *p, server_rec *s, conn_rec *c, request_rec *r,
 #endif
         }
         else if (strcEQ(var, "REMOTE_ADDR"))
+#if AP_SERVER_MINORVERSION_NUMBER <= 2
+            result = c->remote_ip;
+#else
             result = c->client_ip;
+#endif
         else if (strcEQ(var, "HTTPS")) {
             if (sslconn && sslconn->ssl)
                 result = "on";
@@ -307,7 +311,7 @@ static char *nss_var_lookup_ssl(apr_pool_t *p, conn_rec *c, char *var)
     else if (ssl != NULL && strcEQ(var, "SESSION_ID")) {
         char *idstr;
         SECItem *iditem;
- 
+
         if ((iditem = SSL_GetSessionID(ssl)) == NULL)
             return NULL;
 
@@ -345,6 +349,20 @@ static char *nss_var_lookup_ssl(apr_pool_t *p, conn_rec *c, char *var)
             CERT_DestroyCertificate(xs);
         }
     }
+    else if (ssl != NULL && strcEQ(var, "TLS_SNI")) {
+        SECItem *hostInfo = SSL_GetNegotiatedHostInfo(ssl);
+        if (hostInfo) {
+            result = apr_pstrndup(p, (char *) hostInfo->data, hostInfo->len);
+            PORT_Free(hostInfo);
+        }
+    }
+    else if (ssl != NULL && strcEQ(var, "SECURE_RENEG")) {
+        PRInt32 flag = 0;
+#ifdef SSL_ENABLE_RENEGOTIATION
+        SSL_OptionGet(ssl, SSL_ENABLE_RENEGOTIATION, &flag);
+#endif
+        result = apr_pstrdup(p, flag ? "true" : "false");
+    }
 
     return result;
 }
@@ -384,7 +402,7 @@ static char *nss_var_lookup_nss_cert(apr_pool_t *p, CERTCertificate *xs, char *v
     else if (strcEQ(var, "S_DN")) {
         xsname = CERT_NameToAscii(&xs->subject);
         result = apr_pstrdup(p, xsname);
-        PR_Free(xsname);
+        PORT_Free(xsname);
         resdup = FALSE;
     }
     else if (strlen(var) > 5 && strcEQn(var, "S_DN_", 5)) {
@@ -394,7 +412,7 @@ static char *nss_var_lookup_nss_cert(apr_pool_t *p, CERTCertificate *xs, char *v
     else if (strcEQ(var, "I_DN")) {
         xsname = CERT_NameToAscii(&xs->issuer);
         result = apr_pstrdup(p, xsname);
-        PR_Free(xsname);
+        PORT_Free(xsname);
         resdup = FALSE;
     }
     else if (strlen(var) > 5 && strcEQn(var, "I_DN_", 5)) {
@@ -414,7 +432,7 @@ static char *nss_var_lookup_nss_cert(apr_pool_t *p, CERTCertificate *xs, char *v
                 &suite, sizeof suite) == SECSuccess)
             {
                 result = apr_psprintf(p, "%s-%s", suite.macAlgorithmName, suite.authAlgorithmName);
-            } 
+            }
         } else
             result = apr_pstrdup(p, "UNKNOWN");
         resdup = FALSE;
@@ -509,7 +527,7 @@ static char *nss_var_lookup_nss_cert_valid(apr_pool_t *p, CERTCertificate *xs, i
 }
 
 /* Return a string giving the number of days remaining until the cert
- * expires "0" if this can't be determined. 
+ * expires "0" if this can't be determined.
  *
  * In mod_ssl this is more generic, passing in a time to calculate against,
  * but I see no point in converting the end date into a string and back again.
@@ -578,22 +596,17 @@ static char *nss_var_lookup_nss_cert_PEM(apr_pool_t *p, CERTCertificate *xs)
      * similar to mod_ssl. */
     i=0;
     len = strlen(tmp);
-    while (tmp[i] != '\0') {
+    for (i=0; i < len; i++) {
         if (tmp[i] == '\r') {
-            memmove(&tmp[i], &tmp[i+1], 1+(len - i));
+            memmove(&tmp[i], &tmp[i+1], 1+(len - i - 1));
         }
         i++;
     }
 
-    /* Allocate the size of the cert + header + footer + 1 */
-    result = apr_palloc(p, strlen(tmp) + 29 + 27 + 1);
-    strcpy(result, CERT_HEADER);
-    strcat(result, tmp);
-    strcat(result, CERT_TRAILER);
-    result[strlen(tmp) + 29 + 27] = '\0';
+    result = apr_pstrcat(p, CERT_HEADER, tmp, CERT_TRAILER, NULL);
 
     /* Clean up memory. */
-    PR_Free(tmp);
+    PORT_Free(tmp);
 
     return result;
 }
@@ -635,7 +648,7 @@ static char *nss_var_lookup_nss_cert_verify(apr_pool_t *p, conn_rec *c)
 
 static char *nss_var_lookup_nss_cipher(apr_pool_t *p, conn_rec *c, char *var)
 {
-    SSLConnRec *sslconn = myConnConfig(c);    
+    SSLConnRec *sslconn = myConnConfig(c);
     char *result;
     BOOL resdup;
     PRFileDesc *ssl;
@@ -696,8 +709,9 @@ static char *nss_var_lookup_nss_cipher(apr_pool_t *p, conn_rec *c, char *var)
     if (result != NULL && resdup)
         result = apr_pstrdup(p, result);
 
-    PR_Free(issuer);
-    PR_Free(subject);
+    PORT_Free(issuer);
+    PORT_Free(subject);
+    PORT_Free(cipher);
 
     return result;
 }
@@ -724,7 +738,7 @@ static char *nss_var_lookup_nss_version(apr_pool_t *p, char *var)
     return result;
 }
 
-static char *nss_var_lookup_protocol_version(apr_pool_t *p, conn_rec *c) 
+static char *nss_var_lookup_protocol_version(apr_pool_t *p, conn_rec *c)
 {
     char *result;
     SSLChannelInfo      channel;
@@ -799,7 +813,7 @@ void nss_var_log_config_register(apr_pool_t *p)
 static const char *nss_var_log_handler_c(request_rec *r, char *a)
 {
     SSLConnRec *sslconn = myConnConfig(r->connection);
-    char *result; 
+    char *result;
 
     if (sslconn == NULL || sslconn->ssl == NULL)
         return NULL;
